/*ldir/lread
  Program to read Linux extended 2 filesystems under DOS

  Module dir.c
  This file contains routines for handling directories.

  Copyright information and copying policy see file README.TXT

  History see file MAIN.C
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include "ext2.h"
#include "proto.h"
#include "ldir.h"

#define S_IFMT   0x0F000  /*017 0000*/

#define S_IFSOCK 0x0C000  /*014 0000*/
#define S_IFLNK  0x0A000  /*012 0000*/
#define S_IFFIL  0x08000  /*010 0000*/
#define S_IFBLK  0x06000  /*006 0000*/
#define S_IFDIR  0x04000  /*004 0000*/
#define S_IFCHR  0x02000  /*002 0000*/
#define S_IFIFO  0x01000  /*001 0000*/

#define S_ISSOCK(m)     (((m) & S_IFMT) == S_IFSOCK)
#define S_ISLNK(m)      (((m) & S_IFMT) == S_IFLNK)
#define S_ISFIL(m)      (((m) & S_IFMT) == S_IFFIL)
#define S_ISBLK(m)      (((m) & S_IFMT) == S_IFBLK)
#define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)
#define S_ISCHR(m)      (((m) & S_IFMT) == S_IFCHR)
#define S_ISFIFO(m)     (((m) & S_IFMT) == S_IFIFO)

/*not yet supported:*/
#define S_IFUID  0x00800  /*000 4000*/
#define S_IFGID  0x00400  /*000 2000*/
#define S_IFVTX  0x00200  /*000 1000*/


extern super_block sb;
extern enum {LDIR, LREAD} modus;
enum {SDIR,SFILE} source_type = SDIR;

unsigned int allocations=0;    /*for debugging memory allocation*/

void *MALLOC(size_t size)
{ void *p=NULL;
  if (size)
    { p=malloc(size);
#ifdef DEBUG_ALLOC
      printf("malloc: #=%2u  size=%8u  p=%10lp\n",++allocations,(unsigned int) size,p);
#endif
    }
  return(p);
}
void FREE(void *block)
{ if (block!=NULL)
    { free(block);
#ifdef DEBUG_ALLOC
      printf("free:   #=%2u                 p=%10lp\n",allocations--,block);
#endif
    }
}

unsigned long eatpath( char *path, char *inode_name, char ctrl)
{ inode *i;
  unsigned long inode_num, inode_num_old;
  char *path_seg, *temp, *temp1;
  static char sym_path[256];

#ifdef DEBUG_DIR
  printf("----------Executing eatpath------------------------\n");
#endif
start:
  if (!strcmp( path, "/" ))        /* Look for path = root dir */
    { inode_num=(unsigned long) 2;
      strcpy(inode_name,"/");
      return(inode_num);           /* bail quick and early if so */
    }
  if (path[0]=='/')
      strcpy(sym_path,path);
  else
    { strcpy(sym_path,"/");
      strcat(sym_path,path);
    }
  strcpy(path,sym_path);
  if (path[strlen(path)-1]=='/')
      source_type=SDIR;
  else
      source_type=SFILE;

  path_seg = strtok( path, "/" ); /* break out first path element */

  /* Start with root directory */
  inode_num=    (unsigned long) 2;
  inode_num_old=(unsigned long) 2;

  while (1)
    { i = load_inode( inode_num );  /*read in directory*/

    /* If this node is a symlink, we have to restart eatpath() with the
	 absolute path of the linked file */
      if (S_ISLNK(i->i_mode))
	{ fprintf(STDERR,"SYMLINK found -> %s\n",(char *) &i->i_block[0]);
	  if (i->i_size<64)
	    { temp1= (char *) &i->i_block[0];
	      if (temp1[0]=='/')  /*it's an absolute path*/
		{ strcpy (sym_path,temp1);
		}
	      else if ((temp1[0]=='.') && (temp1[1]=='.') && (temp1[2]=='/'))
		{ temp=strrchr(sym_path,'/');   /*it's a path with ../ */
		  if (temp!=NULL) temp[0]='\0';
		  temp=strrchr(sym_path,'/');   /*it's a path with ../ */
		  if (temp!=NULL) temp[1]='\0';
		  strcat(sym_path,(char *) &temp1[3]);
		}
	      else if ((temp1[0]=='.') && (temp1[1]=='/'))
		{ temp=strrchr(sym_path,'/');    /*it's a path with ./ */
		  if (temp!=NULL) temp[1]='\0';
		  strcat(sym_path,(char *) &temp1[2]);
		}
	      else  /*it's a relative path, concatenate path and link*/
		{ temp=strrchr(sym_path,'/');
		  if (temp!=NULL) temp[1]='\0';
		  strcat(sym_path,(char *) &temp1[0]);
		}
	      path=sym_path;
	      goto start;
	    }
	  else
	    { fprintf( STDERR, "Symbolic link '%s' too long, cannot be processed.\n", path_seg );
	      exit( -1 );
	    }
	}

      /*to catch symlinks, we have to read one inode beyond path_seg==NULL,
	therefore the break statement is in the middle of the while-loop*/
      if (path_seg==NULL)
	{ if (modus==LDIR)
	    { if (source_type==SFILE)
		{ inode_num=inode_num_old;
		}
	      break;
	    }
	  else if (modus==LREAD)
	    { inode_num=inode_num;
	      if (S_ISDIR(i->i_mode))  /*file inodes should not be copied*/
		{ fprintf(STDERR,"%s is a directory, cannot be read\n",sym_path);
		  exit(-1);
		}
	      break;
	    }
	}
      inode_num_old=inode_num;
      inode_num    =search_dir( i, path_seg, inode_name, ctrl);

#ifdef DEBUG_DIR
      printf("Inode for %s = %ld  name=%s count=%i\n",path_seg,inode_num,inode_name,count);
#endif
      if (inode_num==0)
	{ if (ctrl==INIT)
	      fprintf( STDERR, "Can't find '%s'.\n", path_seg );
	  return(inode_num);
	}

      /* See if there's more string to parse */
      path_seg = strtok( NULL, "/" ); /* Get next section of path */
    }
  return(inode_num);
}

/*compares strings, returns 0, if strings are equal, accepts wildcards*/
int comp_name(char *filename, char *name)
{ int i,j;

  for (i=0,j=0;i<=strlen(filename);i++,j++) /*compare including \0*/
     { if (filename[i]=='?')
	 { continue;                /*wildcard '?' -> skip this character*/
	 }
       else if (filename[i]=='*')    /*wildcard '*' -> rest of string does not matter*/
	 { if (modus==LREAD)
	     { if (!strcmp(name,".")||!strcmp(name,".."))
		 { return(1);
		 }
	     }
	   for (;filename[i]!='.';i++)
	     { if (filename[i]=='\0')
		   return(0);	     /*strings are equal*/
	     }
	   for (;name[j]!='.';j++)
	     { if (name[j]=='\0')
		   return(1);       /*strings are equal*/
	     }
	 }
       else if (filename[i]!=name[j])
	 { return(1);                /*character not equal -> string not equal*/
	 }
     }
  return(0);                         /*strings are equal*/
}

/* List a directory (names and inode numbers) with inode struct as input */
void list_dir( inode *i , char *path)
{ char  *name, *buf;
  char filename[256];
  inode *in;
  unsigned long c_time, inode_num;
  unsigned int  rec_len, name_len, count = 0;
  size_t size;

#ifdef DEBUG_DIR
  printf("----------Executing list_dir------------------------\n");
  print_inode(i, 0);
#endif

  if (source_type==SFILE) /*if it's a file to list, break out filename without path*/
    { strcpy(filename,path);
      name=strtok(filename,"/");
      while (name)
	{ buf=name;
	  name=strtok(NULL,"/");
	}
      strcpy(filename,buf);
    }

  size=(size_t) ((i->i_blocks) * 512);

  if ((buf = (char *) MALLOC(size)) == NULL )
    { fprintf( STDERR, "Memory problem in list_dir --%s--%s--%s--\n",filename,buf,path );
      exit( -1 );
    }
  if (read_inode( i, buf, size, RELEASE_BUF) == -1 )
    { fprintf(STDERR,"Error with read_inode in dir.c\n");
      exit( -1 );
    }

  while ((count + 8) < size ) /* Don't run over end */
    { inode_num = *((unsigned long  *) &buf[count]);
      rec_len =   *((unsigned short *) &buf[count+4]);
      name_len =  *((unsigned short *) &buf[count+6]);

#ifdef DEBUG_DIR
      printf("\nsize=%i count=%i inode_num=%lu rec_len=%i name_len=%i\n",size,count,inode_num,rec_len,name_len);
#endif

      if ((inode_num<1) || (inode_num>sb.s_inodes_count))
	  break;
      if (inode_num == 0)       /* This means a deleted entry */
	{ count += rec_len;     /* keep reading... */
	  continue;
	}
      if( (name = (char *) MALLOC( name_len+1 ) ) == NULL )
	{ fprintf( STDERR, "Directory too big: build_dir ran out of memory.\n" );
	  exit( -1 );
	}
      strncpy(name, &buf[count + 8], name_len);
      name[name_len] = '\0';              /* terminate name string */

      in = load_inode( inode_num );
      if (in == NULL )
	{ fprintf(STDERR,"Error with load_inode in dir.c\n");
	  exit( -1 );         /* We hit a problem */
	}
      c_time = in->i_ctime+TIMEZONE;

      if (source_type==SFILE)
	{ if (!comp_name(filename,name))
	    { printf( "%10s %4d %4d %9lu %s %s",
		     build_mode( in ), in->i_uid, in->i_gid, in->i_size,
		     strtok(ctime((time_t *) &c_time), "\n"), name );
	      if (S_ISLNK(in->i_mode))   /*Print fast symbolic links (name < 64 char)*/
		{ if (in->i_size<64)
		    { printf(" -> %s",(char *) &in->i_block[0]);
		    }
		  else
		    { printf("-> Symlink name too long");
		    }
		}
	      printf("\n");
	    }
	}
      else
	{ printf( "%10s %4d %4d %9lu %s %s",
	      build_mode( in ), in->i_uid, in->i_gid, in->i_size,
	      strtok(ctime((time_t *) &c_time), "\n"), name );
	  if (S_ISLNK(in->i_mode))   /*Print fast symbolic links (name < 64 char)*/
	    { if (in->i_size<64)
		{ printf(" -> %s",(char *) &in->i_block[0]);
		}
	      else
		{ printf("-> Symlink name too long");
		}
	    }
	  printf("\n");
	}
      FREE(name);
      count += rec_len;                         /* increment counter */
      if ( (rec_len==0) || ((rec_len>255)&&(rec_len>name_len+16)) )
	    break;  /*End of List, but is this correct ???*/
    }
  FREE(buf);
}

/* Interpret a buf as a directory structure and search for an entry
   return inode_number if found, 0 if not
*/
unsigned long search_dir( inode *i, char *str, char *inode_name, char ctrl)
{ unsigned int  rec_len, name_len;
  unsigned long inode_num=0;
  static char *name=NULL, *buf=NULL, *str_search=NULL;
  static size_t buf_size;
  static unsigned int count=0;

#ifdef DEBUG_DIR
  printf("----------Executing search_dir--------------%s-----%i----\n",str,ctrl);
#endif

  if (ctrl==INIT)
    { count=0;
      FREE(buf);
      FREE(str_search);
      if ((buf = (byte *)MALLOC( (i->i_blocks) * 512 )) == NULL )
	{ fprintf( STDERR, "Memory problem in search_dir.\n" );
	  exit( -1 );
	}
      if (read_inode( i, buf, (i->i_blocks) * 512, RELEASE_BUF ) == -1 )
	{ fprintf(STDERR,"Problem in search_dir\n");
	  exit( -1 );
	}
      buf_size= i->i_blocks * 512;
      str_search=strdup(str);
    }
  else if (ctrl==KILL_BUF)
    { FREE(buf);
      FREE(str_search);
      return(0);
    }

  while (count + 8 < buf_size ) /* Don't run over end */
    { inode_num =    *( (unsigned long *)  &buf[count]   );
      rec_len =      *( (unsigned short *) &buf[count+4] );
      name_len =     *( (unsigned short *) &buf[count+6] );

      if ((inode_num<1) || (inode_num>sb.s_inodes_count))
	{ inode_num=0;
	  break;
	}

      if (inode_num == 0)    /* This means a deleted entry */
	{ count += rec_len;     /* keep reading... */
	  continue;
	}
      if( (name = (char *) MALLOC( name_len+1 ) ) == NULL )
	{ fprintf( STDERR, "Directory too big: search_dir_new ran out of memory.\n" );
	  exit( -1 );
	}
      strncpy(name, &buf[count + 8], name_len);
      name[name_len] = '\0';              /* terminate name string */
      count += rec_len;                   /* increment counter */
      if (!comp_name(str_search,name))    /* Do these match? */
	{ strcpy(inode_name,name);
	  FREE(name);                     /* if yes don't search the rest*/
	  break;
	}
      FREE(name);
      if ( (rec_len==0) || ((rec_len>255)&&(rec_len>name_len+16)) )
	{  count=0;
	   inode_num=0;
	   break;
	}
      inode_num=0;
    }
  return(inode_num);
}

char *build_mode( inode *i )
{ static char modestr[11];
  char fullstr[11];
  int x;

#ifdef DEBUG_DIR
  printf("----------Executing build_mode------------------------\n");
#endif

  strcpy( modestr, "?---------" );
  strcpy( fullstr, "-rwxrwxrwx" );

  for( x = 0; x < 9; x++ )
    if( ( i->i_mode >> x ) & 1 )
      modestr[9-x] = fullstr[9-x];

  /* here order counts right now */
  if (S_ISLNK(i->i_mode))
	  modestr[0] = 'l';
  else if (S_ISDIR(i->i_mode))
	  modestr[0] = 'd';
  else if (S_ISFIL(i->i_mode))
	  modestr[0] = '-';
  else if (S_ISCHR(i->i_mode))
	  modestr[0] = 'c';
  else if (S_ISBLK(i->i_mode))
	  modestr[0] = 'b';
  else if (S_ISSOCK(i->i_mode))
	  modestr[0]= 's';
  else if (S_ISFIFO(i->i_mode))
	  modestr[0]= 'p';
/* #ifdef DEBUG_DIR */
  else 
	  printf("Unknown Inode Mode %X\n",i->i_mode);
/* #endif */

  return modestr;

}
