/*ldir/lread
  Program to read Linux extended 2 filesystems under DOS

  Main Program  main.c

  Copyright information and copying policy see file README.TXT

  V1.0: Written by Jason Hunter and Dave Lutz.
  V2.0: Rewritten by Werner Zimmermann, FHTE Esslingen, Germany, Oct. 1996
	-new user interface, new command line options
	-new drive detection code
	-debugged buffering, memory allocation to handle big directories
	-error handling
  V2.0a:Integrated floppy disk support. Full support for (short) symlinks.
	W.Zimmermann, Nov. 9, 96
  V2.0b:Completely rewrote memory allocation stuff in dir.c,
	improved harddisk partition detection.
	W.Zimmermann, Nov. 16,96
  V2.0c:Now also single files can be ldir'ed.
	(Limited) support for wildcards, see README.TXT.
	W.Zimmermann, Nov. 21,96
  V2.0d:lread can now copy multiple files.
	W.Zimmermann, Dec. 6,96
  V2.0e: Initialized inode_num in search_dir to fix bug when doing
	multiple read of files.
	W.Zimmermann, Dec. 11,96
  V2.1: Dynamic buffer allocation in function readdisk in readdisk.c,
	removed READ_BUF_SIZE.
	W.Zimmermann, Jan. 24,97
  V2.2: SCSI support has now also been tested. Some comments modified.
	Now under Windows95 also long filenames for lread are supported
        (unfortunately this does not seem to work with DJPP's gcc)
        W.Zimmermann, May 10, 97
  V2.3: Fixed partition table read. Could not handle multiple linux
        partitions in DOS extended partitions. Bug found by Jakob
        Saternus. Thanks, Jakob!
        Included better Windows 95 detection, now can distinguish between
        pure DOS7 and Win95. Thanks to Santiago Garcia Mantinan!
        W.Zimmermann, June 6, 97
*/

#define VERSION  "V2.3"

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>       /* for tolower() */
#ifdef __DOS__
#include <sys\stat.h>
#include <io.h>          /* for creat/write/close */
#include <dos.h>
#else
#include <sys/stat.h>
#endif
#include <string.h>
#include "ext2.h"
#include "proto.h"
#include "ldir.h"

unsigned int disk_no;
unsigned int part_no;
char *ext2_disk;
enum {LDIR, LREAD} modus = LDIR;

#ifdef __TURBOC__
int dos_long_creat(char *filename)
{ int fd;
  union REGS regs; struct SREGS sregs;

  if (_osmajor>6)
    { regs.x.ax = 0x1600;      /* is Windows 95 installed? */
      int86(0x2F,&regs,&regs);
      if (regs.h.al)
	{ regs.x.ax = 0x716C;  /* if yes, long filenames are supported */
	}
      else
	{ regs.x.ax = 0x6C00;  /* Dos 7.0 */
	}
    }
  else
      regs.x.ax = 0x6C00;  /* Dos 7.0 */

  regs.x.bx = 0x0002;
  regs.x.cx = 0x0000;
  regs.x.dx = 0x0012;
  regs.x.si = FP_OFF(filename);
  sregs.ds  = FP_SEG(filename);

  fd=intdosx(&regs, &regs, &sregs);

  /* if carry flag is set, there was an error */
  if (regs.x.cflag)
      return(-1);
  else
      return(fd);
}
#endif

int main( int argc, char **argv )
{ inode *in;
  byte   buf[1024];
  int    bytes, fd, x;
  unsigned long inode_num=0;
  char   inode_name[256], sourcebak[256], *source, *destin, *temp, ans,
	 destin_mode=NORM;

  /* change the argv[0] to lowercase */
  for (x=0; x < strlen( argv[0] ); x++)
     argv[0][x] = tolower( argv[0][x] );

  if (strstr(argv[0],"ldir"))           /* See if its ldir or lread*/
      modus = LDIR;
  else                                  /* it was lread */
      modus = LREAD;

  if ((argc>1) && (strstr(argv[1],"-h")||strstr(argv[1],"-v"))) /* See if help is requested*/
    { fprintf(STDERR,"LDIR/LREAD version %s (C) 1997 W.Zimmermann\n",VERSION);
      fprintf(STDERR,"usage:  ldir  [-s/dev/hd..] /LinuxPath/LinuxDir\n");
      fprintf(STDERR,"  or    lread [-s/dev/hd..] /LinuxPath/LinuxFile [DosFile]\n" );
      fprintf(STDERR,"  or    ldir -h    or lread -h\n");
      fprintf(STDERR,"If you have problems (especially SCSI drives!), read the README file\n");
      exit(-1);
    }

  if ((argc>1) && strstr(argv[1],"-s")) /* See if default disk is used */
    { x=2;
      ext2_disk=&argv[1][2];            /* strip the '-s'*/
    }
  else
    { x=1;
      ext2_disk=LINUX_DISK;             /* default disk*/
    }

  if (strstr(ext2_disk,"/dev/hda"))     /* figure out disk number for DOS*/
      disk_no=128;
  else if (strstr(ext2_disk,"/dev/hdb"))
      disk_no=129;
  else if (strstr(ext2_disk,"/dev/hdc"))
      disk_no=130;                          /*correct number???*/
  else if (strstr(ext2_disk,"/dev/hdd"))
      disk_no=131;                          /*correct number???*/
  else if (strstr(ext2_disk,"/dev/sda"))
      disk_no=128;                          /*correct number???*/
  else if (strstr(ext2_disk,"/dev/sdb"))
      disk_no=129;                          /*correct number???*/
  else if (strstr(ext2_disk,"/dev/fd0"))
      disk_no=0;
  else if (strstr(ext2_disk,"/dev/fd1"))
      disk_no=1;
  else
    { printf("Unknown Drive Specification: %s   %i\n",ext2_disk,disk_no);
      exit(-1);
    }
  if (disk_no>=128)  /*do it only for harddisks ???*/
    { strcpy(sourcebak,ext2_disk);              /*figure out partition number*/
      if (strstr(ext2_disk,"dev/hd"))
	{ temp=strtok(sourcebak,"/dev/hd");
	}
      else
	{ temp=strtok(sourcebak,"dev/sd");
          fprintf(STDERR,"SCSI drive support is experimental, see the README file\n");
	}
      part_no=temp[1]-0x30;                     /*convert char to int*/
      if ((part_no<1)||(part_no>9))
	{ part_no=0;
	}
    }
  else
      part_no=1; /*on floppy disks we only have one partition*/

  if ((argc==1)&&(modus==LREAD))
    { fprintf(STDERR,"You must specify a Source Filename (help: lread -h)\n");
      exit(-1);
    }
  if (argc>x)
    { source=argv[x];
      if (argc>x+1)
	{ if (!strcmp(argv[x+1],"."))       /*copy destin to same filename as source?*/
	      destin_mode=SELF;
	  else
	      destin=strdup(argv[x+1]);
	}
      else
	  destin="STDOUT";
    }
  else
    { source="/";
      destin="STDOUT";
    }

#ifdef __TURBOC__
  _fmode = O_BINARY;   /* we want all files binary for this */
#endif

#ifdef __DOS__
  examine_drive();
#else
  open_dev(ext2_disk);
#endif

  if (load_super() == -1)                 /* Couldn't get superblock */
    { printf("Error Loading Superblock\n");
      exit( -1 );
    }
  if (load_groups() == -1)                /* Couldn't read groups */
    { printf("Error Reading Groups\n");
      exit( -1 );
    }
#ifdef DEBUG_MAIN
  printf("Printing Group--------------------------\n");
  print_groups();
#endif
  strcpy( sourcebak, source );
  inode_num=eatpath( source, inode_name,INIT);
  if (inode_num == 0)                  /* Linux path was invalid */
    { printf("Linux Path Invalid\n");
      exit(-1);
    }

#ifdef DEBUG_MAIN
  printf("argc:%i---x:%i---Modus:%i---Source:%s---Destination:%s---\n",argc,x,modus,sourcebak,destin);
  printf("Printing inode------------------------------------\n");
  print_inode( in, inode_num );
  printf("Printing block-------------------------------------\n");
  print_blocks( in );
#endif
  switch (modus)                        /* See if a dir is requested */
   { case LDIR: { search_dir( NULL, NULL, inode_name, KILL_BUF);
		  in = load_inode( inode_num );
		  if (in == NULL)
		    { printf("Inode Error\n");
		      exit(-1);
		    }
		  list_dir( in, sourcebak );
		  exit( 0 );
		}
     case LREAD:{ while (inode_num)
		    { if (destin_mode==SELF)
			  destin=inode_name;
		      in = load_inode( inode_num );
		      if (in == NULL)                         /* Problem with the inode */
			{ printf("Inode Error\n");
			  exit(-1);
			}
		      if (strcmp(destin,"STDOUT")) /*See if output to a file*/
			{ if (fopen(destin,"r"))
			    { fprintf(STDERR,"File %s exists, overwrite? [Y/N]  ",destin);
			      scanf("%c",&ans);
			      if ((char) toupper(ans)!='Y')
				{ fprintf(STDERR,"Could not copy file %s to %s\n",source,destin);
				  exit(-1);
				}
			    }
#ifdef __TURBOC__
			  fd=dos_long_creat(destin);
#else
			  fd=creat( destin, S_IREAD | S_IWRITE );
#endif
			  if (fd == -1)
			    { printf("Problem opening second argument file (DOS file) %s %s\n",sourcebak,destin);
			      exit( -1 );
			    }
			}
		      else           /*we want output to stdout */
			{ fd = 1;
			}

		      if ( (bytes=read_inode(in, buf, 1024, KEEP_BUF ))==-1)  /* read first block */
			{ fprintf(STDERR, "Error Reading Inode\n");
			  exit(0);
			}
		      do
			{ if (write(fd,buf,bytes)!=bytes)   /* write to disk */
			    { fprintf(STDERR,"Error Writing File %s\n",destin);
			      exit(-1);
			    }
			  bytes = read_inode( NULL, buf, 1024, KEEP_BUF );
			  if (bytes == -1)          /* problem reading */
			    { printf("Error Reading 2\n");
			      exit (-1);
			    }
			} while( bytes != 0 );
		      if (bytes)
			  printf( "*** Problem encountered reading file %s***\n",sourcebak);
		      if (fd>2)
			  close( fd );
		      read_inode(NULL,NULL,0,KILL_BUF); /*release the buffer*/
		      strcpy(source,sourcebak);
		      inode_num=search_dir( NULL, NULL, inode_name, KEEP_BUF);
		    }
		  search_dir( NULL, NULL, inode_name, KILL_BUF);
		}
   }
  return(0);
}
