/*ldir/lread
  Program to read Linux extended 2 filesystems under DOS

  Module readdisk.c
  Low level harddisk partition table and harddisk data read
  If we are under DOS, install the first half of the file.
  Otherwise, install the second half.

  Copyright information and copying policy see file README.TXT

  History see file MAIN.C
*/

#ifdef __DOS__

/* This file's primary purpose is to give DOS the ability to read
 * a set of bytes from the Linux partition.  It has to look at the
 * partition table and find where the Linux partition is and what
 * the drive's parameters look like so it can convert a ulong
 * representative of a logical byte number in the partition to
 * a track #, sector #, head assignment, and offset in that sector.
 * It's in essence a DOS version of Linux's lseek off /dev/hd...
*/

#include <stdio.h>
#include <bios.h>
#include <dos.h>
#include <stdlib.h>
#include <conio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "ldir.h"

extern void *MALLOC(size_t size);
extern void FREE(void *block);

#define  PRINT_PARTITION_TABLE    for (k=0;k<4;k++)                     \
				     { printf("Entry %i: ",k);          \
				       for (i=0;i<16;i++)               \
					    printf("%2x  ",run[k][i]);  \
				       printf("\n");                    \
				     }                                  \
				  printf("Partition Code %x%x\n",buf[510],buf[511]);

/* globals */
unsigned int  HEADS;                    /*your harddisk's # of heads               ) drive's */
unsigned int  SECTORS;                  /*                # of sectors per cylinder) geometry*/
unsigned int  CYLINDERS;                /*                # of cylinders           )         */
unsigned long start;                    /*logical block address(LBA) of your Linux partition*/
unsigned long num_sect;                 /*total # of sectors of your Linux partition*/
extern unsigned int  disk_no;           /*DOS' disk #, eg. 0x80=first harddisk*/
extern unsigned int  part_no;           /*# of your Linux partition*/
extern char *ext2_disk;                 /*your linux partition name, eg. /dev/hda5*/

/*Convert Logical Block Number (LBA) adress to
  Cylinder-Head-Sector (CHS) address
  DOS and Linux software doesn't care about heads, sectors, cylinders, ie.
  your harddisk's hardware details and therefore works with LBA, but if you
  want to access your harddisk via the BIOS, you have to call the BIOS rou-
  tines with CHS addresses.
*/
void convert(unsigned long x, unsigned int *head, unsigned int *track, \
	     unsigned int *sector, unsigned int *offset)
{ unsigned long logicsect, absolsect;

#ifdef DEBUG_READDISK
  printf("--------Executing 'convert'----------------\n");
#endif
  logicsect =  x >> 9;                 /*divide by 512= harddisk sector size*/
  *offset   =  x & 0x1FF;              /*modulo 512*/
  absolsect =  logicsect+start;

  *track    =  absolsect / (SECTORS * HEADS);
  *head     = (absolsect / SECTORS) % HEADS;
  *sector   = (absolsect % SECTORS) + 1;

#ifdef DEBUG_READDISK
  printf ("Beginning at sect=%ld...\n",start);
  printf ("byte=%4ld logic: %-4ld absol: %-4ld head: %-2d track: %-3d sect: %-4d  offset: %-d\n",
		x,logicsect,absolsect,*head,*track,*sector,*offset);
#endif
  if (absolsect > start+num_sect)
    { fprintf(STDERR,"You cannot access ABOVE the Linux partition.\n");
      exit(-1);
    }
}

/*This procedure looks for your Linux partition.
  You have to specify the disk to search in ldir.h or via the -s command
  line switch, specification is done in Linux 'style', your specification is
  converted into a DOS 'style' specification in global variables disk_no and
  part_no, eg.:

      /dev/hdaX           first harddisk        disk_no=0x80
      /dev/hdbX           second harddisk               0x81
      /dev/fd0            first floppy disk             0x00
      /dev/fd1            second floppy disk            0x01

  If you do not specify a partition number, i.e. if X is a space, the procedure
  will search the four entrys of the partition table for a primary Linux Ext2
  partition, if it finds one, it will use it. If it does not find a primary
  Linux Ext2 partition, it searches for a DOS extended partition, read's the
  DOS extended partition's table and searches it for a Linux Ext2 partition.
  As DOS extended partitions may contain an unlimited number of 'logical drives',
  this search is recursively until we find a Linux Ext2 partition or until we
  reach the end of the tables.

  If you additionally specify a partition number, i.e. set X to 5, for each
  Linux partition we find we also check, if it is the specified partition num-
  ber. If not, we continue the search of the partition table. The partition
  number is transfered to the procedure via global variable part_no.
*/

void examine_drive(void)
{ int i,j=0,k,kfound,p=1,ptemp[8];
  static unsigned char *buf;
  char found=NONE;
  unsigned char run[4][16];

  unsigned int head=0, cyl=0, sect=1;
  unsigned int first_head, first_sect, first_cyl;

#ifdef DEBUG_READDISK
  unsigned int boot, end_head, end_sect, end_cyl;
  unsigned int system;

  printf("--------Executing 'examine drive'----------------\n");
#endif

  buf=(unsigned char *) MALLOC(512);

  biosdisk(PARA_CMD,disk_no,head,cyl,sect,1,buf);  /*let BIOS tell us the drive's geometry*/
  HEADS=buf[3]+1;
  CYLINDERS=buf[1]+(buf[0]>>6)*256+1;
  SECTORS=buf[0] & 0x3F;

#ifdef DEBUG_READDISK
  printf("Search for %s   disk=%i  part=%i  heads=%i   sectors=%i   cylinders=%i\n",
	  ext2_disk,disk_no,part_no,HEADS,SECTORS,CYLINDERS);
#endif

  if ((disk_no==0)||(disk_no==1))       /*Is it a floppy drive? 0=A:  1=B:*/
    { start     = (long) 0;             /*floppy's do not have partitions*/
      SECTORS   = SECTORS;              /*all data starts at LBA 0*/
      num_sect  = HEADS*SECTORS*CYLINDERS;
#ifdef DEBUG_READDISK
      printf("heads=%i  sectors=%i  cylinders=%i num_sect:%i\n", HEADS, SECTORS, CYLINDERS,num_sect);
#endif
      FREE(buf);
      return;
    }

  /* read the partition table from the first hard drive, partition table starts at CHS=0 0 1*/
  if (biosdisk (READ_CMD,disk_no,head,cyl,sect,1,buf))
    { fprintf(STDERR,"Error Reading Partition Table on %s\n",ext2_disk);
      fprintf(STDERR,"Use command line switch -s/dev/hdx[y] , where x=a,b,... is the\n");
      fprintf(STDERR,"Linux harddisk and y=1,2,... (may be omitted) is your partition number\n");
      fprintf(STDERR,"Example: ldir -s/dev/hdb6\n");
      fprintf(STDERR,"To avoid this message, edit ldir.h and recompile (see README.TXT)\n");
      fprintf(STDERR,"If you use a mix of IDE and SCSI disks, you definitely should read README.TXT\n");
      exit (-1);  /*exit, if the BIOS can't read the partition table*/
    }
#ifdef DEBUG_READDISK
  printf("------------------------\nPrimary Partition Table at Head:%i  Cylinder:%i   Sector:%i Partition:1-4 \n",head,cyl,sect);
#endif
  for (k=0;k<4;k++)             /*Search the Primary Partition Table for a Linux partition*/
    { for (i=0;i<16;i++)        /*The partition table has 4 entrys with 16 bytes per entry*/
	  run[k][i] = buf[446+(k*16)+i]; /*it starts at byte 446 in the harddisk's first sector*/
      if (run[k][4]==EXT2PART)  /*EXT2PART=0x83 is the Linux Ext2 partition signature*/
	{
#ifdef DEBUG_READDISK
	  printf("Linux Partition Found        at Entry#=%i Part#=%i  ",k,k+1);
#endif
	  if ((part_no==0)||(k==part_no-1))  /*we found a Linux partition, but does it have the correct partition number?*/
	    { found=LINUX;
	      kfound=k;                      /*yes*/
#ifdef DEBUG_READDISK
	      printf("->it's the right one\n");
#endif
	      break;
	    }
	  else
            { ptemp[j++]=k+1;                /*no*/
#ifdef DEBUG_READDISK
	      printf("->it's NOT the right one\n");
#endif
	    }
	}
    }
#ifdef DEBUG_READDISK
  PRINT_PARTITION_TABLE;
#endif

  if (found!=LINUX)             /*We haven't found a primary Linux partition, so we search for an extended partition*/
    { for (k=0;k<4;k++)
	  if (run[k][4]==0x05)  /*0x05 is the signature for extended partitions*/
	    { found=EXTENDED;
	      p=5;              /*Extended partitions start at /dev/hd?5*/
	      kfound=k;
	    }
    }
  if (found==EXTENDED)          /*Searching extended partition table*/
    do
      { head= run[kfound][1];   /*find out the CHS-address of the extended partition's partition table*/
	sect= run[kfound][2];
	cyl = run[kfound][3];
	cyl+= (sect>>6)*256;    /* add top 2 bits of first_sect */
	sect= sect & 0x3F;      /* and remove them from first_sect */
	if (biosdisk (READ_CMD,disk_no,head,cyl,sect,1,buf)&0xff)         /*read it*/
	  { fprintf(STDERR,"Error Reading Partition Table\n");
	    exit(-1);
	  }
#ifdef DEBUG_READDISK
	printf("------------------------\nSecondary Partition Table at Head:%i  Cylinder:%i   Sector:%i Partition:%i \n",head,cyl,sect,p);
#endif
	for (k=0;k<4;k++)       /*Check extended Partition Table, same game as with the primary table*/
	   { for (i=0;i<16;i++) /*The extended partition table has 2 entrys with 16 bytes per entry*/
		 run[k][i] = buf[446+(k*16)+i]; /*where one entry can be another extended partition*/
	     if (run[k][4]==EXT2PART)            /*hurrah, it's Linux*/
	       {
#ifdef DEBUG_READDISK
		 printf("Linux Partition Found        at Entry#=%i Part#=%i  ",k,p);
#endif
		 if ((part_no==0)||(p==part_no)) /*does it have the correct partition number?*/
		   { found=LINUX;
		     kfound=k;                   /*yes*/
#ifdef DEBUG_READDISK
		     printf("->it's the right one\n");
#endif
		     break;
		   }
		 else
                   { ptemp[j++]=p;               /*no*/
#ifdef DEBUG_READDISK
		     printf("->it's NOT the right one\n");
#endif
		   }
	       }
	     if (run[k][4]==0x05)
	       { found=EXTENDED;            /*hurrah, another extended partion*/
		 kfound=k;
		 p++;
#ifdef DEBUG_READDISK
		 printf("DOS Extended Partition Found at Entry#=%i Part#=%i\n",kfound,p);
#endif
	       }
	   }
#ifdef DEBUG_READDISK
	PRINT_PARTITION_TABLE;
#endif
      } while ((run[1][4]!=0)   && (found!=LINUX) &&
	       (buf[510]==0x55) && (buf[511]==0xAA));
/*our search continues until either -we found the last entry which's signature is 0
			     or     -we found the right Linux Ext2 partition
			     or     -the sector is not a partition table (signature 0xAA55*/

  if (found!=LINUX)   /*if we finally did not find the Linux Ext2 partition we exit*/
    { fprintf(STDERR,"Could not find Linux partition on %s\n",ext2_disk);
      if (j>0)
          for(i=0;i<j;i++)
              fprintf(STDERR,"Found Linux at partition number %i\n",ptemp[i]);
      fprintf(STDERR,"Use command line switch -s/dev/hdx[y] , where x=a,b,... is the\n");
      fprintf(STDERR,"Linux harddisk and y=1,2,... (may be omitted) is your partition number\n");
      fprintf(STDERR,"Example: ldir -s/dev/hdb6\n");
      fprintf(STDERR,"To avoid this message, edit ldir.h and recompile (see README.TXT)\n");
      fprintf(STDERR,"If you use a mix of IDE and SCSI disks, you definitely should read README.TXT\n");
      exit (-1);
    }

  /*if we reach this point, we've found the right Linux Ext2 partition*/
  first_head=run[kfound][1];  /*get the partitions CHS address from the partition table*/
  first_sect=run[kfound][2];
  first_cyl =run[kfound][3];
  first_cyl+=(first_sect>>6)*256;   /* add top 2 bits of first_sect */
  first_sect= first_sect & 0x3F;    /* and remove them from first_sect */

  start     = (long) first_sect + ((long) first_head)*SECTORS \
	      + ((long) first_cyl)*(HEADS*SECTORS) - 1;
  num_sect  =(long) run[k][15];
  num_sect  = num_sect*256 + run[kfound][14];
  num_sect  = num_sect*256 + run[kfound][13];
  num_sect  = num_sect*256 + run[kfound][12];

  /* Brag about the results */
#ifdef DEBUG_READDISK
  boot      =run[kfound][0];
  system    =run[kfound][4];
  end_head  =run[kfound][5];
  end_sect  =run[kfound][6];
  end_cyl   =run[kfound][7];
  end_cyl  +=(end_sect>>6)*256;      /* add top 2 bits of end_sect */
  end_sect  = end_sect & 0x3F;       /* and remove them from end_sect */

  printf ("\n      THE DRIVE HAS:\n");
  printf ("    number of heads: %d\n", HEADS);
  printf ("  number of sectors: %d\n", SECTORS);
  printf ("number of cylinders: %d\n", CYLINDERS);
  printf ("\n   LINUX ITSELF HAS:\n");
  printf ("     boot indicator: %xH\n",boot);
  printf ("     beginning head: %d\n", first_head);
  printf ("   beginning sector: %d\n", first_sect);
  printf (" beginning cylinder: %d\n", first_cyl);
  printf ("   system indicator: %xH\n",system);
  printf ("        ending head: %d\n", end_head);
  printf ("      ending sector: %d\n", end_sect);
  printf ("    ending cylinder: %d\n", end_cyl);
  printf ("relative start sect: %ld\n",start);
  printf ("  number of sectors: %ld\n",num_sect);
#endif
  FREE(buf);
}



int readdisk( unsigned char *buf, unsigned long loc, size_t size )
{
  unsigned int head,cyl,sect,offset;
  unsigned int rc, x;
  unsigned long no_sect;
  static unsigned char *temp;

#ifdef DEBUG_READDISK
  printf("--------Executing 'read disk'--*buf:%u--size:%u--loc:%lu\n",*buf,size,loc);
#endif

  convert (loc,&head,&cyl,&sect,&offset);

  /*With floppys there are problems when trying to read beyond the last sector
    of a cylinder. So we try to deal with that. I never had this problem with
    harddisks, most likely, because my harddisks work in lba-mode, translating
    heads/sector/cylinder on their own. If we get read errors (rd!=0), the
    following piece of code needs to be improved
  */
  no_sect= size >> 9;
  if (disk_no<128)
    { if (no_sect<1)
	  no_sect=no_sect+1;    /*with floppys: minimum read is 1 sector*/
    }
  else
    { no_sect=no_sect+2;        /*with harddisks:               2 sectors*/
    }

  temp=(unsigned char *) MALLOC(no_sect*512); /*???*/

  for (x=0;x<3;x++)  /*for floppy disks we sometimes have to try more than once*/
    {  rc= biosdisk (READ_CMD,disk_no,head,cyl,sect,no_sect,temp);
       if (rc==0) break;
    }

  if (rc==0)
    { memcpy(buf,&temp[offset],size);  /* no ending NULL */
      FREE(temp);
      return size;
    }
  else
    { fprintf(STDERR,"Error %x in Read Disk - loc:%lu head:%u cyl:%u sect:%u  no_sect:%lu\n",rc,loc,head,cyl,sect,num_sect);
      FREE(temp);
      return 0;
    }
}


/* If we're under Linux (for testing) include the following */

#else

/* Read Linux */
/* Written by Dave Lutz */

/* This file is to be used for testing this software under Linux. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int IN;                         /* holds the open input dev */
typedef unsigned char byte;

int open_dev( char *name )
{
  return( IN = open( name, O_RDONLY ) ); /* open the file */
}

int readdisk( unsigned char *buf, unsigned long loc, size_t size )
{
  int i;
#ifdef DEBUG_READDISK
  printf("--------Executing 'read disk'--*buf:%u--size:%u--loc:%lu\n",*buf,size,loc);
#endif

  lseek( IN, loc, SEEK_SET );
  i=read(IN, buf, size );
  return i;
}

#endif
